 (setf snd-head-none 0) 

 (setf snd-head-AIFF 1) 

 (setf snd-head-IRCAM 2) 

 (setf snd-head-NeXT 3) 

 (setf snd-head-Wave 4) 

 (setf snd-head-PAF 5) 

 (setf snd-head-SVX 6) 

 (setf snd-head-NIST 7) 

 (setf snd-head-VOC 8) 

 (setf snd-head-W64 9) 

 (setf snd-head-MAT4 10) 

 (setf snd-head-MAT5 11) 

 (setf snd-head-PVF 12) 

 (setf snd-head-XI 13) 

 (setf snd-head-HTK 14) 

 (setf snd-head-SDS 15) 

 (setf snd-head-AVR 16) 

 (setf snd-head-SD2 17) 

 (setf snd-head-FLAC 18) 

 (setf snd-head-CAF 19) 

 (setf snd-head-raw 20) 
 
 (setf snd-head-OGG 21) 

 (setf snd-head-channels 1) 

 (setf snd-head-mode 2) 

 (setf snd-head-bits 4) 

 (setf snd-head-srate 8) 

 (setf snd-head-dur 16) 

 (setf snd-head-latency 32) 

 (setf snd-head-type 64) 

 (setf snd-mode-adpcm 0) 

 (setf snd-mode-pcm 1) 

 (setf snd-mode-ulaw 2) 

 (setf snd-mode-alaw 3) 

 (setf snd-mode-float 4) 

 (setf snd-mode-upcm 5) 

 (setf snd-mode-unknown 6) 

 (setf snd-mode-double 7) 

 (setf snd-mode-GSM610 8) 

 (setf snd-mode-DWVW 9) 

 (setf snd-mode-DPCM 10) 

 (setf snd-mode-msadpcm 11) 
 
 (setf snd-mode-vorbis 12)

 (SETF MAX-STOP-TIME 10E20) 

 (SETF MIN-START-TIME -10E20) 

 (setf OP-AVERAGE 1) (setf OP-PEAK 2) 

